from typing import Union, Optional, Callable, TYPE_CHECKING
from pathlib import Path
from deltalake import DeltaTable

if TYPE_CHECKING:
    from azure.core.credentials import TokenCredential


def get_delta_table(
    uri_or_table: Union[str, Path, DeltaTable],
    storage_options: Optional[dict] = None,
    *,
    get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]" = None,
) -> DeltaTable:
    base_path = (
        uri_or_table.table_uri
        if isinstance(uri_or_table, DeltaTable)
        else (
            uri_or_table
            if isinstance(uri_or_table, str)
            else str(uri_or_table.absolute())
        )
    )
    base_path = base_path.removesuffix("/")
    if isinstance(uri_or_table, Path) or isinstance(uri_or_table, str):
        from .azure_helper import get_storage_options_object_store

        path_for_delta, storage_options_for_delta = get_storage_options_object_store(
            uri_or_table, storage_options, get_credential
        )
        return DeltaTable(path_for_delta, storage_options=storage_options_for_delta)
    return uri_or_table



from pathlib import Path
from typing import TYPE_CHECKING, cast, Union, Optional, Callable

from deltalake2db.azure_helper import get_storage_options_object_store
from deltalake2db.filter_by_meta import _can_filter

if TYPE_CHECKING:
    import polars as pl
    from azure.core.credentials import TokenCredential
from deltalake import DeltaTable, Field, DataType
from deltalake.schema import StructType, ArrayType, MapType
from collections import OrderedDict
import os


def _get_expr(
    base_expr: "Union[pl.Expr,None]",
    dtype: "DataType",
    meta: Optional[Field],
    parquet_field: "Optional[pl.PolarsDataType]",
) -> "pl.Expr":
    import polars as pl

    if parquet_field is None:
        return (
            pl.lit(None, _get_type(dtype)).alias(meta.name)
            if meta
            else pl.lit(None, _get_type(dtype))
        )

    pn = (
        meta.metadata.get("delta.columnMapping.physicalName", meta.name)
        if meta
        else None
    )
    if base_expr is None:
        assert pn is not None
        base_expr = pl.col(pn)
    if isinstance(dtype, StructType):

        def _get_sub_type(name: str) -> "pl.PolarsDataType| None":
            if parquet_field is None:
                return None
            assert isinstance(parquet_field, pl.Struct)
            for f in parquet_field.fields:
                if f.name == name:
                    return f.dtype
            return None

        struct_vl = (
            pl.when(base_expr.is_null())
            .then(pl.lit(None))
            .otherwise(
                pl.struct(
                    *[
                        _get_expr(
                            base_expr.struct.field(
                                subfield.metadata.get(
                                    "delta.columnMapping.physicalName", subfield.name
                                )
                            ),
                            subfield.type,
                            subfield,
                            _get_sub_type(
                                subfield.metadata.get(
                                    "delta.columnMapping.physicalName", subfield.name
                                )
                            ),
                        )
                        for subfield in dtype.fields
                    ]
                )
            )
        )
        return struct_vl.alias(meta.name) if meta else struct_vl
    elif isinstance(dtype, ArrayType):
        list_vl = (
            pl.when(base_expr.is_null())
            .then(pl.lit(None))
            .otherwise(
                base_expr.list.eval(
                    _get_expr(
                        pl.element(),
                        dtype=dtype.element_type,
                        meta=meta,
                        parquet_field=cast(pl.List, parquet_field).inner,
                    )
                )
            )
        )
        return list_vl.alias(meta.name) if meta else list_vl
    return base_expr.alias(meta.name) if meta else base_expr


def _get_type(dtype: "DataType") -> "pl.PolarsDataType":
    import polars as pl

    if isinstance(dtype, StructType):
        return pl.Struct([pl.Field(f.name, _get_type(f.type)) for f in dtype.fields])
    if isinstance(dtype, ArrayType):
        return pl.List(_get_type(dtype.element_type))
    if isinstance(dtype, MapType):
        raise NotImplementedError("MapType not supported in polars")

    dtype_str = str(dtype.type)
    if dtype_str == "string":
        return pl.String
    elif dtype_str in ["int", "integer"]:
        return pl.Int32
    elif dtype_str == "long":
        return pl.Int64
    elif dtype_str == "double":
        return pl.Float64
    elif dtype_str == "float":
        return pl.Float32
    elif dtype_str == "boolean":
        return pl.Boolean
    elif dtype_str == "date":
        return pl.Date
    elif dtype_str == "timestamp":
        return pl.Datetime
    elif dtype_str == "binary":
        return pl.Binary
    elif dtype_str.startswith("decimal"):
        precision, scale = dtype_str.split("(")[1].split(")")[0].split(",")
        return pl.Decimal(int(precision), int(scale))
    elif dtype_str == "short":
        return pl.Int16
    elif dtype == "byte":
        return pl.Int8
    elif dtype_str == "null":
        return pl.Null
    elif dtype_str in ["timestampNtz", "timestamp_ntz"]:
        return pl.Datetime
    raise NotImplementedError(f"{dtype_str} not supported in polars currently")


def _filter_cond(f: "pl.LazyFrame", conditions: dict) -> "pl.LazyFrame":
    return f.filter(**conditions)


def get_polars_schema(
    delta_table: Union[DeltaTable, Path, str],
) -> "OrderedDict[str, pl.PolarsDataType]":
    from .protocol_check import check_is_supported

    if isinstance(delta_table, Path) or isinstance(delta_table, str):
        delta_table = DeltaTable(delta_table)
    check_is_supported(delta_table)
    res_dict = OrderedDict()
    for f in delta_table.schema().fields:
        res_dict[f.name] = _get_type(f.type)
    return res_dict


def scan_delta_union(
    delta_table: Union[DeltaTable, Path, str],
    conditions: Optional[dict] = None,
    storage_options: Optional[dict] = None,
    *,
    get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]" = None,
) -> "pl.LazyFrame":
    import polars as pl
    from .protocol_check import check_is_supported

    if isinstance(delta_table, Path) or isinstance(delta_table, str):
        path_for_delta, storage_options_for_delta = get_storage_options_object_store(
            delta_table, storage_options, get_credential
        )
        delta_table = DeltaTable(
            path_for_delta, storage_options=storage_options_for_delta
        )
    check_is_supported(delta_table)
    all_ds = []
    all_fields = delta_table.schema().fields
    for ac in delta_table.get_add_actions(flatten=True).to_pylist():
        if conditions is not None and _can_filter(ac, conditions):
            continue
        fullpath = os.path.join(delta_table.table_uri, ac["path"]).replace("\\", "/")
        base_ds = pl.scan_parquet(
            fullpath, storage_options=delta_table._storage_options, glob=False
        )
        try:
            parquet_schema = base_ds.limit(0).collect_schema()
        except AttributeError:
            parquet_schema = base_ds.limit(0).schema  # old polars
        selects = []
        for field in all_fields:
            pl_dtype = _get_type(field.type)
            pn = field.metadata.get("delta.columnMapping.physicalName", field.name)
            if "partition_values" in ac and pn in ac["partition_values"]:
                part_vl = ac["partition_values"][pn]
                selects.append(pl.lit(part_vl, pl_dtype).alias(field.name))
            elif "partition." + pn in ac:
                part_vl = ac["partition." + pn]
                selects.append(pl.lit(part_vl, pl_dtype).alias(field.name))
            elif "partition_values" in ac and field.name in ac["partition_values"]:
                # as of delta 0.14
                part_vl = ac["partition_values"][field.name]
                selects.append(pl.lit(part_vl, pl_dtype).alias(field.name))
            elif "partition." + field.name in ac:
                # as of delta 0.14
                part_vl = ac["partition." + field.name]
                selects.append(pl.lit(part_vl, pl_dtype).alias(field.name))
            elif (
                field.metadata.get("delta.columnMapping.physicalName", field.name)
                in parquet_schema
            ):
                selects.append(
                    _get_expr(
                        None,
                        field.type,
                        field,
                        parquet_schema.get(
                            field.metadata.get(
                                "delta.columnMapping.physicalName", field.name
                            )
                        ),
                    )
                )
            else:
                selects.append(pl.lit(None, pl_dtype).alias(field.name))
        ds = (
            _filter_cond(base_ds.select(*selects), conditions)
            if conditions is not None
            else base_ds.select(*selects)
        )
        all_ds.append(ds)
    if len(all_ds) == 0:
        return pl.DataFrame(
            data=[], schema={f.name: _get_type(f.type) for f in all_fields}
        ).lazy()
    return pl.concat(all_ds, how="diagonal_relaxed")



import logging

logger = logging.getLogger(__name__)


def _can_filter(action: dict, conditions: dict):
    # partition.COLNAME
    # min.COLNAME
    # max.COLNAME
    try:
        for key, value in conditions.items():
            part_vl = action.get("partition." + key, None)
            if part_vl is not None and part_vl != value:
                return True
            min_vl = action.get("min." + key, None)
            max_vl = action.get("max." + key, None)
            if isinstance(min_vl, str):
                value = value[0 : len(min_vl)]
            if (
                min_vl is not None
                and max_vl is not None
                and (value < min_vl or value > max_vl)
            ):
                return True
        return False
    except Exception as e:
        logger.warning(f"Could not filter: {e}")
        return False



def get_az_blob_fs(fake_proto: str, **kwargs):
    import adlfs

    class AzureBlobFS(adlfs.AzureBlobFileSystem):
        protocol = fake_proto

        def __init__(self, real_proto, *args, **kwargs):
            super().__init__(*args, **kwargs)

        @classmethod
        def _strip_protocol(cls, path: str):
            path = path.replace(fake_proto, "abfs")
            return super()._strip_protocol(path)

    return AzureBlobFS("abfs", **kwargs)



from deltalake import DeltaTable
from deltalake.exceptions import DeltaProtocolError


supported_reader_features = [
    "columnMapping",
    "timestampNtz",
]  # not: deletionVectors, v2Checkpoint


def is_protocol_supported(dt: DeltaTable):
    prot = dt.protocol()
    if prot.min_reader_version <= 3:
        return True
    assert prot.reader_features is not None
    un_supported = [
        f for f in prot.reader_features if f not in supported_reader_features
    ]
    return not any(un_supported)


def check_is_supported(dt: DeltaTable):
    prot = dt.protocol()
    if prot.min_reader_version <= 3:
        return
    assert prot.reader_features is not None
    un_supported = [
        f for f in prot.reader_features if f not in supported_reader_features
    ]
    if not any(un_supported):
        return
    raise DeltaProtocolError(
        f"Delta table features not supported: {', '.join(un_supported)}"
    )



from .duckdb import get_sql_for_delta
from .duckdb import get_sql_for_delta_expr
from .duckdb import create_view_for_delta as duckdb_create_view_for_delta
from .duckdb import apply_storage_options as duckdb_apply_storage_options
from .polars import scan_delta_union as polars_scan_delta, get_polars_schema
from .protocol_check import is_protocol_supported
from .delta_lake import get_delta_table

__all__ = [
    "get_sql_for_delta",
    "get_sql_for_delta_expr",
    "duckdb_create_view_for_delta",
    "duckdb_apply_storage_options",
    "polars_scan_delta",
    "get_polars_schema",
    "is_protocol_supported",
    "get_delta_table",
]



from pathlib import Path
from typing import Optional, Sequence, Any, TypeVar
import sqlglot.expressions as ex
from typing import Union


def read_parquet(
    path: Union[str, Path, list[Path], list[str], ex.Expression, list[ex.Expression]],
) -> ex.Expression:
    if isinstance(path, list):
        return ex.func(
            "read_parquet",
            ex.array(
                *[
                    ex.Literal.string(str(p)) if isinstance(p, (str, Path)) else p
                    for p in path
                ]
            ),
            ex.EQ(
                this=ex.Column(this=ex.Identifier(this="union_by_name", quoted=False)),
                expression=ex.Boolean(this=True),
            ),
        )
    return ex.func(
        "read_parquet",
        ex.Literal.string(str(path)) if isinstance(path, (str, Path)) else path,
    )


def union(selects: Sequence[ex.Expression], *, distinct: bool) -> ex.Expression:
    if len(selects) == 0:
        raise ValueError("No selects to union")
    elif len(selects) == 1:
        return selects[0]
    elif len(selects) == 2:
        return ex.union(selects[0], selects[1], distinct=distinct)
    else:
        return ex.union(
            selects[0], union(selects[1:], distinct=distinct), distinct=distinct
        )


def filter_via_dict(conditions: Optional[dict[str, Any]]):
    if not conditions or len(conditions) == 0:
        return None
    return [
        (
            ex.EQ(this=ex.column(k, quoted=True), expression=ex.convert(v))
            if v is not None
            else ex.Is(this=ex.column(k, quoted=True), expression=ex.Null())
        )
        for k, v in conditions.items()
    ]


T = TypeVar("T", bound=ex.Query)


def merge_ctes(select: T, other_select: ex.Select) -> T:
    new_select = select.copy()
    for cte in other_select.ctes:
        new_select.with_(cte.alias, cte.this, copy=False)
    return new_select


def struct(items: dict[str, ex.Expression]) -> ex.Struct:
    return ex.Struct(
        expressions=[
            ex.PropertyEQ(this=ex.Identifier(this=k, quoted=True), expression=v)
            for k, v in items.items()
        ]
    )


def list_transform(
    param_names: Union[list[str], str],
    list_expr: ex.Expression,
    value_expr: ex.Expression,
) -> ex.Expression:
    if isinstance(param_names, str):
        param_names = [param_names]
    return ex.func(
        "list_transform",
        list_expr,
        ex.Lambda(
            this=value_expr,
            expressions=[ex.Identifier(this=p, quoted=False) for p in param_names],
        ),
    )



from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Optional, Union
from deltalake import DataType, Field
from deltalake.schema import StructType, ArrayType, PrimitiveType, MapType
import sqlglot.expressions as ex
from deltalake2db.azure_helper import (
    get_storage_options_fsspec,
    get_storage_options_object_store,
    AZURE_EMULATOR_CONNECTION_STRING,
)
from deltalake2db.filter_by_meta import _can_filter
import deltalake2db.sql_utils as squ

if TYPE_CHECKING:
    from deltalake import DeltaTable
    from azure.core.credentials import TokenCredential
    import duckdb


def _cast(s: ex.Expression, t: Optional[ex.DATA_TYPE]):
    if t is None:
        return s
    return ex.cast(s, t)


def _dummy_expr(
    field_type: "PrimitiveType | StructType | ArrayType | MapType",
) -> ex.Expression:
    from deltalake.schema import PrimitiveType

    if isinstance(field_type, PrimitiveType):
        if str(field_type).startswith("decimal("):
            cast_as = ex.DataType.build(str(field_type))
        else:
            cast_as = type_map.get(field_type.type)
        return _cast(ex.Null(), cast_as)
    elif isinstance(field_type, StructType):
        return squ.struct(
            {
                subfield.name: _dummy_expr(subfield.type)
                for subfield in field_type.fields
            }
        )
    elif isinstance(field_type, ArrayType):
        return ex.Array(expressions=[_dummy_expr(field_type.element_type)])
    elif isinstance(field_type, MapType):
        return ex.MapFromEntries(
            this=ex.Array(
                expressions=[
                    ex.Tuple(
                        expressions=[
                            _dummy_expr(field_type.key_type),
                            _dummy_expr(field_type.value_type),
                        ]
                    )
                ]
            )
        )
    raise ValueError(f"Unsupported type {field_type}")


def _get_expr(
    base_expr: "ex.Expression|None",
    dtype: "DataType",
    meta: Optional[Field],
    alias=True,
    *,
    counter: int = 0,
) -> "ex.Expression":
    pn = (
        meta.metadata.get("delta.columnMapping.physicalName", meta.name)
        if meta
        else None
    )
    if base_expr is None:
        assert pn is not None
        base_expr = ex.Column(this=pn)
    if isinstance(dtype, StructType):
        struct_expr = (
            ex.case()
            .when(base_expr.is_(ex.Null()), ex.Null())
            .else_(
                squ.struct(
                    {
                        subfield.name: _get_expr(
                            ex.Dot(
                                this=base_expr,
                                expression=ex.Identifier(
                                    this=subfield.metadata.get(
                                        "delta.columnMapping.physicalName",
                                        subfield.name,
                                    ),
                                    quoted=True,
                                ),
                            ),
                            subfield.type,
                            subfield,
                            alias=False,
                            counter=counter + 1,
                        )
                        for subfield in dtype.fields
                    }
                )
            )
        )
        if alias and meta is not None:
            return struct_expr.as_(meta.name)
        return struct_expr
    elif isinstance(dtype, ArrayType):
        vl = squ.list_transform(
            "x_" + str(counter),
            base_expr,
            _get_expr(
                ex.Identifier(this="x_" + str(counter), quoted=False),
                dtype.element_type,
                meta=meta,
                alias=False,
                counter=counter + 1,
            ),
        )
        if alias and meta is not None:
            return vl.as_(meta.name)
        return vl

    return base_expr.as_(meta.name) if meta is not None and alias else base_expr


def load_install_azure(con: "duckdb.DuckDBPyConnection"):
    with con.cursor() as cur:
        cur.execute(
            "select loaded, installed from duckdb_extensions() where extension_name='azure' "
        )
        res = cur.fetchone()
        loaded = res[0] if res else False
        installed = res[0] if res else False
    if not installed:
        con.install_extension("azure")
    if not loaded:
        con.load_extension("azure")


def apply_storage_options(
    con: "duckdb.DuckDBPyConnection",
    uri: Union[str, Path],
    storage_options: Optional[dict],
    *,
    use_fsspec: bool = False,
):
    if storage_options is None:
        return
    if isinstance(uri, str) and (
        ".blob.core.windows.net" in uri or ".dfs.core.windows.net" in uri
    ):
        from urllib.parse import urlparse

        up = urlparse(uri)
        account_name_from_url = up.netloc.split(".")[0]
    else:
        account_name_from_url = None
    if use_fsspec:
        apply_storage_options_fsspec(
            con,
            uri if isinstance(uri, str) else str(uri.absolute()),
            storage_options,
            account_name_from_url,
        )
    else:
        apply_storage_options_azure_ext(
            con,
            storage_options,
            account_name_path=account_name_from_url,
        )


def apply_storage_options_fsspec(
    con: "duckdb.DuckDBPyConnection",
    base_path: str,
    storage_options: dict,
    account_name_path: Optional[str] = None,
):
    use_emulator = storage_options.get("use_emulator", "0") in ["1", "True", "true"]

    opts = get_storage_options_fsspec(storage_options)
    account_name = storage_options.get(
        "account_name",
        storage_options.get("azure_storage_account_name", account_name_path),
    )
    if not use_emulator:
        assert account_name is not None, "account_name must be provided for fsspec"
    proto_name_duckdb = "fsspec_az_" + (account_name or "_emulator")
    if not con.filesystem_is_registered(proto_name_duckdb):
        from .fsspec_fs import get_az_blob_fs

        fs = get_az_blob_fs(proto_name_duckdb, **opts)

        con.register_filesystem(fs)  # type: ignore

    return proto_name_duckdb


def apply_storage_options_azure_ext(
    con: "duckdb.DuckDBPyConnection",
    storage_options: dict,
    *,
    type="azure",
    account_name_path: Optional[str] = None,
):
    if type in ["az", "abfs", "abfss"]:
        type = "azure"
    if type != "azure":
        raise ValueError("Only azure is supported for now")
    load_install_azure(con)
    with con.cursor() as cur:
        cur.execute("FROM duckdb_secrets() where type='azure';")
        secrets = cur.fetchall()
        se_names = [s[0] for s in secrets]
    if len(se_names) > 0:  # for now, only one secret is supported in DuckDB, it seems
        return
    account_name = storage_options.get(
        "account_name",
        storage_options.get("azure_storage_account_name", account_name_path),
    )
    use_emulator = storage_options.get("use_emulator", "0") in ["1", "True", "true"]
    secret_name = account_name or ("_emulator" if use_emulator else "")

    if secret_name not in se_names:
        if "connection_string" in storage_options:
            cr = storage_options["connection_string"]
            con.execute(
                f"""CREATE SECRET {secret_name} (
TYPE AZURE,
CONNECTION_STRING '{cr}'
);"""
            )
        elif account_name is not None and "account_key" in storage_options:
            ak = storage_options["account_key"]
            conn_str = f"AccountName={account_name};AccountKey={ak};BlobEndpoint=https://{account_name}.blob.core.windows.net;"
            con.execute(
                f"""CREATE SECRET {secret_name} (
TYPE AZURE,
CONNECTION_STRING '{conn_str}',
ACCOUNT_NAME '{account_name}'
);"""
            )
        elif account_name is not None and "sas_token" in storage_options:
            sas_token = storage_options["sas_token"]
            conn_str = f"BlobEndpoint=https://{account_name}.blob.core.windows.net;SharedAccessSignature={sas_token}"
            con.execute(
                f"""CREATE SECRET {secret_name} (
TYPE AZURE,
CONNECTION_STRING '{conn_str}',
ACCOUNT_NAME '{account_name}'
);"""
            )
        elif account_name is not None and "client_id" in storage_options:
            con.execute(
                f"""CREATE SECRET {secret_name} (
TYPE AZURE,
PROVIDER SERVICE_PRINCIPAL,
TENANT_ID  '{storage_options.get("tenant_id")}',
CLIENT_ID   '{storage_options.get("client_id")}',
CLIENT_SECRET    '{storage_options.get("client_secret")}',
ACCOUNT_NAME '{account_name}'
);"""
            )
        elif account_name is not None:
            chain = storage_options.get("chain", "default")
            con.execute(
                f"""CREATE SECRET {secret_name} (
TYPE AZURE,
PROVIDER CREDENTIAL_CHAIN,
CHAIN '{chain}',
ACCOUNT_NAME '{account_name}'
);"""
            )
        elif use_emulator:
            con.execute(
                f"""CREATE SECRET {secret_name} (
TYPE AZURE,
CONNECTION_STRING '{AZURE_EMULATOR_CONNECTION_STRING}',
ACCOUNT_NAME 'devstoreaccount1'
);"""
            )
        else:
            raise ValueError(
                "No connection string or account_name and account_key provided"
            )
    return secret_name


type_map = {
    "byte": ex.DataType.Type.TINYINT,
    "int": ex.DataType.Type.INT,
    "integer": ex.DataType.Type.INT,
    "long": ex.DataType.Type.BIGINT,
    "boolean": ex.DataType.Type.BOOLEAN,
    "date": ex.DataType.Type.DATE,
    "timestamp": ex.DataType.Type.TIMESTAMPTZ,
    "float": ex.DataType.Type.FLOAT,
    "double": ex.DataType.Type.DOUBLE,
    "string": ex.DataType.Type.VARCHAR,
    "short": ex.DataType.Type.SMALLINT,
    "binary": ex.DataType.Type.BINARY,
    "timestampNtz": ex.DataType.Type.TIMESTAMP,
    "timestamp_ntz": ex.DataType.Type.TIMESTAMP,
    "decimal": ex.DataType.Type.DECIMAL,
}


def create_view_for_delta(
    con: "duckdb.DuckDBPyConnection",
    delta_table: "Union[DeltaTable , Path , str]",
    view_name: str,
    overwrite=True,
    *,
    conditions: Optional[dict] = None,
    storage_options: Optional[dict] = None,
    get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]" = None,
    use_fsspec: bool = False,
):
    sql = get_sql_for_delta(
        delta_table,
        duck_con=con,
        conditions=conditions,
        storage_options=storage_options,
        get_credential=get_credential,
        use_fsspec=use_fsspec,
    )
    assert '"' not in view_name
    if overwrite:
        con.execute(f'create or replace view "{view_name}" as {sql}')
    else:
        con.execute(f'create view "{view_name}" as {sql}')


def get_sql_for_delta_expr(
    dt: "Union[DeltaTable,Path , str]",
    conditions: Union[Optional[dict], Sequence[ex.Expression], ex.Expression] = None,
    select: Union[Sequence[Union[str, ex.Expression]], None] = None,
    distinct=False,
    cte_wrap_name: Union[str, None] = None,
    action_filter: Union[Callable[[dict], bool], None] = None,
    sql_prefix="delta",
    delta_table_cte_name: Union[str, None] = None,
    duck_con: "Union[duckdb.DuckDBPyConnection, None]" = None,
    storage_options: Optional[dict] = None,
    *,
    get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]" = None,
    use_fsspec=False,
) -> ex.Select:
    from deltalake import DeltaTable
    from .sql_utils import read_parquet, union, filter_via_dict

    account_name_path = None
    base_path = (
        dt.table_uri
        if isinstance(dt, DeltaTable)
        else (dt if isinstance(dt, str) else str(dt.absolute()))
    )
    base_path = base_path.removesuffix("/")
    if isinstance(dt, Path) or isinstance(dt, str):
        path_for_delta, storage_options_for_delta = get_storage_options_object_store(
            dt, storage_options, get_credential
        )
        account_name_path = (
            storage_options_for_delta.get("account_name", None)
            if storage_options_for_delta
            else None
        )
        dt = DeltaTable(path_for_delta, storage_options=storage_options_for_delta)

    from .protocol_check import check_is_supported

    check_is_supported(dt)

    delta_table_cte_name = delta_table_cte_name or sql_prefix + "_delta_table"
    from deltalake.schema import PrimitiveType

    file_selects: list[ex.Select] = []

    delta_fields = dt.schema().fields
    owns_con = False
    if duck_con is None:
        import duckdb

        duck_con = duckdb.connect()
        owns_con = True
    if use_fsspec:
        fake_protocol = apply_storage_options_fsspec(
            duck_con,
            base_path,
            storage_options or dt._storage_options or {},
            account_name_path=account_name_path,
        )
        base_path = fake_protocol + "://" + base_path.split("://")[1]
    elif dt.table_uri.startswith("az://") or dt.table_uri.startswith("abfss://"):
        apply_storage_options_azure_ext(
            duck_con,
            storage_options or dt._storage_options or {},
            type="azure",
            account_name_path=account_name_path,
        )  # type: ignore

    try:
        for ac in dt.get_add_actions(flatten=True).to_pylist():
            if (
                conditions is not None
                and isinstance(conditions, dict)
                and _can_filter(ac, conditions)
            ):
                continue
            if action_filter and not action_filter(ac):
                continue
            fullpath = base_path + "/" + ac["path"]
            with duck_con.cursor() as cur:
                cur.execute(f"select name from parquet_schema('{fullpath}')")
                cols: list[str] = [c[0] for c in cur.fetchall()]
            cols_sql: list[ex.Expression] = []
            for field in delta_fields:
                field_name = field.name
                phys_name = field.metadata.get(
                    "delta.columnMapping.physicalName", field_name
                )

                cast_as = None

                if isinstance(field.type, PrimitiveType):
                    if str(field.type).startswith("decimal("):
                        cast_as = ex.DataType.build(str(field.type))
                    else:
                        cast_as = type_map.get(field.type.type)
                if "partition_values" in ac and phys_name in ac["partition_values"]:
                    cols_sql.append(
                        _cast(
                            ex.convert(ac["partition_values"][phys_name]), cast_as
                        ).as_(field_name)
                    )
                elif "partition." + phys_name in ac:
                    cols_sql.append(
                        _cast(ex.convert(ac["partition." + phys_name]), cast_as).as_(
                            field_name
                        )
                    )
                elif "partition_values" in ac and field.name in ac["partition_values"]:
                    cols_sql.append(
                        _cast(
                            ex.convert(ac["partition_values"][field.name]), cast_as
                        ).as_(field_name)
                    )
                elif "partition." + field.name in ac:
                    cols_sql.append(
                        _cast(ex.convert(ac["partition." + field.name]), cast_as).as_(
                            field_name
                        )
                    )
                elif phys_name in cols:
                    cols_sql.append(
                        _get_expr(ex.column(phys_name, quoted=True), field.type, field)
                    )
                else:
                    cols_sql.append(ex.Null().as_(field_name))

            select_pq = ex.select(
                *cols_sql
            ).from_(
                read_parquet(ex.convert(fullpath))
            )  # "SELECT " + ", ".join(cols_sql) + " FROM read_parquet('" + fullpath + "')"
            file_selects.append(select_pq)
        if len(file_selects) == 0:
            file_selects = []
            fields = [_dummy_expr(field.type).as_(field.name) for field in delta_fields]
            file_selects.append(ex.select(*fields).where("1=0"))
        file_sql = ex.CTE(
            this=union(file_selects, distinct=False), alias=delta_table_cte_name
        )
        if select:
            select_exprs = [
                ex.column(s, quoted=True) if isinstance(s, str) else s for s in select
            ]
        else:
            select_exprs = [ex.Star()]

        conds = (
            filter_via_dict(conditions) if isinstance(conditions, dict) else conditions
        )

        se = ex.select(*select_exprs)
        if distinct:
            se = se.distinct()
        if conds is not None:
            se = se.where(*conds)
        se = se.from_(delta_table_cte_name)

        if cte_wrap_name:
            s = ex.select(ex.Star()).from_(cte_wrap_name)
            # s = s.with_()
            s.with_(file_sql.alias, file_sql.args["this"], copy=False)
            s.with_(cte_wrap_name, se, copy=False)
            return s
        else:
            se.with_(file_sql.alias, file_sql.args["this"], copy=False)
            return se
    finally:
        if owns_con:
            duck_con.close()


def get_sql_for_delta(
    dt: "Union[DeltaTable, Path, str]",
    conditions: Optional[dict] = None,
    select: Union[list[str], None] = None,
    distinct=False,
    action_filter: Union[Callable[[dict], bool], None] = None,
    cte_wrap_name: Union[str, None] = None,
    sql_prefix="delta",
    duck_con: "Union[duckdb.DuckDBPyConnection, None]" = None,
    storage_options: Optional[dict] = None,
    *,
    get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]" = None,
    use_fsspec: bool = False,
) -> str:
    expr = get_sql_for_delta_expr(
        dt=dt,
        conditions=conditions,
        select=select,
        distinct=distinct,
        action_filter=action_filter,
        sql_prefix=sql_prefix,
        cte_wrap_name=cte_wrap_name,
        duck_con=duck_con,
        storage_options=storage_options,
        get_credential=get_credential,
        use_fsspec=use_fsspec,
    )
    if cte_wrap_name:
        suffix_sql = ex.select(ex.Star()).from_(cte_wrap_name).sql(dialect="duckdb")
        ws = expr.sql(dialect="duckdb")
        without_suffix = ws.removesuffix(suffix_sql)
        assert without_suffix != ws
        return without_suffix
    return expr.sql(dialect="duckdb")



from typing import Optional, TYPE_CHECKING, Callable, Union
from pathlib import Path

if TYPE_CHECKING:
    from azure.core.credentials import TokenCredential

AZURE_EMULATOR_CONNECTION_STRING = "DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==;BlobEndpoint=http://127.0.0.1:10000/devstoreaccount1;"
STORAGE_SCOPE = "https://storage.azure.com/.default"


def _get_credential_from_chain(
    chain: str, get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]"
):
    from azure.identity import (
        ChainedTokenCredential,
        DefaultAzureCredential,
        ManagedIdentityCredential,
        AzureCliCredential,
        EnvironmentCredential,
    )

    get_credential = get_credential or (lambda x: None)

    creds = chain.split(";")
    map = {
        "cli": AzureCliCredential,
        "env": EnvironmentCredential,
        "managed_identity": ManagedIdentityCredential,
        "default": DefaultAzureCredential,
    }
    creds = [get_credential(c) or map[c]() for c in creds]
    cred = ChainedTokenCredential(*creds) if len(creds) > 1 else creds[0]
    return cred


def get_storage_options_fsspec(storage_options: dict):
    use_emulator = storage_options.get("use_emulator", "0") in ["1", "True", "true"]
    if "connection_string" not in storage_options and use_emulator:
        return {"connection_string": AZURE_EMULATOR_CONNECTION_STRING}
    elif (
        "account_key" not in storage_options
        and "anon" not in storage_options
        and "sas_token" not in storage_options
        and "account_name" in storage_options
        and "chain" not in storage_options
    ):  # anon is true by default in fsspec which makes no sense mostly
        return {"anon": False} | storage_options
    if "chain" in storage_options:
        chain = storage_options["chain"]
        creds = chain.split(";")
        assert len(creds) == 1, "chain must be one credential for fsspec"
        cred = creds[0]
        storage_options2 = storage_options.copy()
        storage_options2.pop("chain")
        if cred == "default":
            return {"anon": False} | storage_options2
        map_flags = {
            "_1": "exclude_workload_identity_credential",
            "_2": "exclude_developer_cli_credential",
            "cli": "exclude_cli_credential",
            "env": "exclude_environment_credential",
            "managed_identity": "exclude_managed_identity_credential",
            "_3": "exclude_powershell_credential",
            "_4": "exclude_visual_studio_code_credential",
            "_5": "exclude_shared_token_cache_credential",
            "_6": "exclude_interactive_browser_credential",
        }
        for k, v in map_flags.items():
            storage_options2[v] = k != cred
        return storage_options2 | {"anon": False}
    return storage_options


def get_storage_options_object_store(
    path: Union[Path, str],
    storage_options: Optional[dict],
    get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]",
):
    # support for "Chain" as in duckdb: https://duckdb.org/docs/extensions/azure#credential_chain-provider
    if storage_options is None:
        return path, None
    new_path = path
    if isinstance(new_path, str) and (
        ".blob.core.windows.net" in new_path or ".dfs.core.windows.net" in new_path
    ):
        from urllib.parse import urlparse

        up = urlparse(new_path)
        account_name_from_url = up.netloc.split(".")[0]

        new_path = new_path.replace(
            up.scheme + "://" + up.netloc, up.scheme + "://"
        ).replace(":///", "://")
    else:
        account_name_from_url = None
    chain = storage_options.get("chain", None)
    get_credential = get_credential or (lambda x: None)

    def _get_cred(chain: str):
        if get_credential is None:
            return None

    if chain is not None:
        new_opts = storage_options.copy()
        new_opts.pop("chain")
        cred = _get_credential_from_chain(chain, get_credential)
        new_opts["token"] = cred.get_token(STORAGE_SCOPE).token
        if account_name_from_url:
            new_opts["account_name"] = new_opts.get(
                "account_name", account_name_from_url
            )
        return new_path, new_opts
    if account_name_from_url is not None and "account_name" not in storage_options:
        new_opts = storage_options.copy()
        new_opts["account_name"] = account_name_from_url
        return new_path, new_opts
    return new_path, storage_options
